using System;
using System.Collections;
using Ringlet.Misc.VCS;

namespace Ringlet.Collections
{
	/// <summary>
	/// An implementation of a tiered vector.
	/// </summary>
	/// <remarks>
	/// <p>A tiered vector is a two-layered data structure characterized
	/// by a numerical value usually referred to as <see cref="n" />.
	/// It consists of a vector of size 2^n containing pointers to 2^n
	/// more vectors of size 2^n, which in their turn contain the actual
	/// sorted keys.</p>
	/// <p>For more information, see <a
	/// href="http://www.ics.uci.edu/~goodrich/pubs/wads99.ps">Tiered
	/// Vectors - Efficient Dynamic Arrays for Rank-Based Sequences</a>,
	/// a paper by Michael T. Goodrich and John G. Kloss II.</p>
	/// </remarks>
	[VCSAttribute("$Ringlet: cs/misc/RingletCollections/Ringlet.Collections/TieredVector.cs,v 1.6 2004/04/09 13:22:33 roam Exp $")]
	public class TieredVector : ICollection, IEnumerable, IList, ICloneable
	{
		/// <summary>
		/// The default characteristic value of a vector.
		/// </summary>
		/// <remarks>This value will be used as <see cref="n" /> if it
		/// is not provided to the TieredVector class constructor.
		/// </remarks>
		public const int TieredVectorDefaultN = 16;

		/// <summary>
		/// The value of N - the tiered vector's characteristic value.
		/// </summary>
		/// <remarks>As noted above, the <see cref="deques" /> array
		/// will consist of 2^N deques, each comprising up to 2^N
		/// actual keys.</remarks>
		protected int n;
		/// <summary>
		/// The value of 2^N - the number of deques within the
		/// vector, the number of elements within a deque.
		/// </summary>
		protected int n2;
		/// <summary>
		/// The value of 2^N-1, cached to facilitate calculations.
		/// </summary>
		protected int mask;
		/// <summary>
		/// The actual array of 2^N deques.
		/// </summary>
		protected RotDeque[] deques;
		/// <summary>
		/// The index of the deque containing the first empty
		/// element.
		/// </summary>
		protected int next;
		
		/// <summary>
		/// Initialize a TieredVector object using the default
		/// characteristic value <see cref="TieredVectorDefaultN" />
		/// </summary>
		public TieredVector()
			: this(TieredVectorDefaultN)
		{
		}

		/// <summary>
		/// Initialize a TieredVector object with a specified N.
		/// </summary>
		/// <param name="nn">The tiered vector characteristic.</param>
		/// <exception cref="System.ArgumentOutOfRangeException">
		/// A value less than 1 was specified for the <paramref
		/// name="nn">nn</paramref> parameter.
		/// </exception>
		public TieredVector(int nn)
		{
			int i;

			if (nn < 1)
				throw (new ArgumentOutOfRangeException(
				    "nn", nn, "The value of N must be an integer > 1"));
			n = nn;
			n2 = 1 << n;
			mask = n2 - 1;
			deques = new RotDeque[n2];
			for (i = 0; i < n2; i++)
				deques[i] = new RotDeque(n2);
		}

		/// <summary>
		/// Get the number of elements in the vector.
		/// </summary>
		public int Count
		{
			get
			{
				return (next * n2 + deques[next].Count);
			}
		}

		/// <summary>
		/// Get the total vector capacity.
		/// </summary>
		public int Capacity
		{
			get
			{
				return (n2 * n2);
			}
		}

		/// <summary>
		/// Check if the vector is empty.
		/// </summary>
		public bool Empty
		{
			get
			{
				return (next == 0 && deques[0].Empty);
			}
		}

		/// <summary>
		/// Check if the vector is filled to capacity and adding
		/// another element would cause it to grow.
		/// </summary>
		public bool Full
		{
			get
			{
				return (next == mask && deques[next].Full);
			}
		}

		/// <summary>
		/// Find the deque in which an object should be placed.
		/// </summary>
		/// <param name="o">The object to examine.</param>
		/// <returns>The index in the deques[] array.</returns>
		protected int GetPosInD(IComparable o)
		{
			int i, j, k;

			// Trivial
			if (next == 0)
				return (0);

			// Binary search
			i = 0;
			j = next + 1;
			while (i + 1< j)
			{
				k = (i + j) / 2;
				if (o.CompareTo(deques[k].AtFront) < 0)
					j = k;
				else
					i = k;
			}

			// Now d[i].AtFront <= o < d[i + 1].AtFront
			return (i);
		}

		/// <summary>
		/// Find the position at which an object would be placed
		/// in a deque.
		/// </summary>
		/// <param name="idx">The deque's index.</param>
		/// <param name="o">The object to examine.</param>
		/// <returns>The position within the specified deque.
		/// </returns>
		protected int GetPosInDeque(int idx, IComparable o)
		{
			RotDeque d;
			int i, j, k;

			d = deques[idx];

			// Trivial
			if (d.Empty)
				return (0);
			if (d.Count == 1)
			{
				if (o.CompareTo(d.AtFront) < 0)
					return (0);
				else
					return (1);
			}
			if (o.CompareTo(d.AtFront) < 0)
				return (0);

			// Binary search
			i = 0;
			j = d.Count;
			while (i + 1 < j)
			{
				k = (i + j) / 2;
				if (o.CompareTo(d[k]) < 0)
					j = k;
				else
					i = k;
			}
			// Now d[i] <= o <= d[i + 1]
			return (i + 1);
		}

		/// <summary>
		/// Adds an element into the vector, preserving the sorting
		/// of the rest.
		/// </summary>
		/// <param name="o">The element to add.</param>
		/// <returns>The position of the newly-added element.
		/// </returns>
		public virtual int Add(object o)
		{
			return (Add((IComparable)o));
		}

		/// <summary>
		/// Removes all elements from the vector.
		/// </summary>
		public virtual void Clear()
		{
			int i;

			for (i = 0; i <= next; i++)
				deques[i].Clear();
		}

		/// <summary>
		/// Determines whether an element is contained in the vector.
		/// </summary>
		/// <param name="o">The element to look for.</param>
		/// <returns>True if the vector contains this element.
		/// </returns>
		public virtual bool Contains(object o)
		{
			return (IndexOf(o) != -1);
		}

		/// <summary>
		/// Find the first occurrence of an element in the vector.
		/// </summary>
		/// <param name="o">The element to look for.</param>
		/// <returns>The index of the first occurrence, or -1 if not
		/// found.</returns>
		public virtual int IndexOf(object o)
		{
			return (IndexOf((IComparable)o));
		}

		/// <summary>
		/// Find the first occurrence of an element in the vector.
		/// </summary>
		/// <param name="o">The element to look for.</param>
		/// <returns>The index of the first occurrence, or -1 if not
		/// found.</returns>
		public virtual int IndexOf(IComparable o)
		{
			int d, p;

			if (Empty)
				return (-1);
			d = GetPosInD(o);
			p = GetPosInDeque(d, o);
			if (p == 0)
				return (-1);
			if (o.CompareTo(deques[d][p - 1]) != 0)
				return (-1);
			return (d * n2 + p - 1);
		}

		/// <summary>
		/// Insert an element at a specified position into the tiered
		/// vector.
		/// </summary>
		/// <param name="pos">The position to insert the element at.
		/// </param>
		/// <param name="o">The element to insert.</param>
		/// <remarks>Unless overridden by a derived class, this method
		/// will always throw an InvalidOperationException, since the
		/// tiered vector does not support direct insertion for fear
		/// of unsorting the elements.
		/// </remarks>
		/// <exception cref="System.InvalidOperationException">
		/// Direct insertion into a tiered vector is not supported.
		/// </exception>
		public virtual void Insert(int pos, object o)
		{
			throw (new InvalidOperationException(
			    "Direct insertion into a tiered vector is not supported"));
		}

		/// <summary>
		/// Get a value indicating whether the tiered vector has a
		/// fixed size.
		/// </summary>
		public virtual bool IsFixedSize
		{
			get
			{
				return (false);
			}
		}

		/// <summary>
		/// Get a value indicating whether the tiered vector is
		/// read-only.
		/// </summary>
		public virtual bool IsReadOnly
		{
			get
			{
				return (false);
			}
		}

		/// <summary>
		/// Remove the first occurrence of the specified object.
		/// </summary>
		/// <param name="o">The object to remove.</param>
		public virtual void Remove(object o)
		{
			Remove((IComparable)o);
		}

		/// <summary>
		/// Remove the first occurrence of the specified object.
		/// </summary>
		/// <param name="o">The object to remove.</param>
		public virtual void Remove(IComparable o)
		{
			int idx;

			idx = IndexOf(o);
			if (idx == -1)
				return;
			RemoveAt(idx);
		}

		/// <summary>
		/// Remove the element at the specified index.
		/// </summary>
		/// <param name="idx">The index of the element to remove.</param>
		/// <exception cref="System.ArgumentOutOfRangeException">
		/// A negative or too large value was specified for the
		/// <paramref name="idx">idx</paramref> parameter.
		/// </exception>
		public virtual void RemoveAt(int idx)
		{
			int d, p, i;

			if (idx < 0 || idx >= Count)
				throw (new ArgumentOutOfRangeException("idx"));
			d = idx >> n;
			p = idx & mask;
			deques[d].RemoveAt(p);
			if (d < next)
			{
				for (i = d; i < next; i++)
					deques[i].PushBack(deques[i + 1].PopFront());
			}
			if (deques[next].Empty && next > 0)
			{
				next--;
				if (next < n2 / 8)
					Shrink();
			}
		}

		/// <summary>
		/// Add an element to the tiered vector.
		/// </summary>
		/// <param name="o">The element to add.</param>
		/// <returns>The position the element was added at.</returns>
		public virtual int Add(IComparable o)
		{
			int d, pos, i;

			// Let's get over this once and for all...
			if (Full)
				Grow();

			d = GetPosInD(o);
			if (deques[d].Empty)
			{
				// Oh.. that was easy! ;)
				deques[d].PushBack(o);
				return (d * n2);
			}
			pos = GetPosInDeque(d, o);
			// Off-by-one?
			if (deques[d].Count == n2 &&
			    o.CompareTo(deques[d].AtBack) >= 0)
			{
				d++;
				pos = 0;
				if (d > next)
					next = d;
			}
			// Free up space if necessary
			if (deques[d].Count == n2)
			{
				if (deques[next].Count == n2)
					next++;
				for (i = next; i > d; i--)
					deques[i].PushFront(deques[i - 1].PopBack());
			}
			deques[d].Insert(pos, o);
			return (d * n2 + pos);
		}

		/// <summary>
		/// Retrieve a vector element by index.
		/// </summary>
		/// <exception cref="System.ArgumentOutOfRangeException">
		/// A negative or overly large value was specified for the
		/// <paramref name="idx">idx</paramref> parameter.
		/// </exception>
		/// <exception cref="System.InvalidOperationException">
		/// Direct setting of tiered vector elements is not supported.
		/// </exception>
		public virtual object this[int idx]
		{
			get
			{
				int d, p;

				d = idx >> n;
				p = idx & mask;
				if (idx < 0 || d > next ||
				    (d == next && p >= deques[next].Count))
					throw (new ArgumentOutOfRangeException("idx", idx,
					   "Invalid element index specified within the tiered vector"));
				return (deques[d][p]);
			}

			set
			{
				throw (new InvalidOperationException(
				    "Direct setting of tiered vector elements is not supported."));
			}
		}

		/// <summary>
		/// Grow the tiered vector.
		/// </summary>
		/// <exception cref="System.InvalidOperationException">
		/// A tiered vector should only be grown when it is full.
		/// </exception>
		protected virtual void Grow()
		{
			int nn, nn2, nmask, i;
			RotDeque[] ndq;

			if (!Full)
				throw (new InvalidOperationException("Cannot grow a non-full tiered vector!"));

			// Create the new variables
			nn = n + 1;
			nn2 = 1 << nn;
			nmask = nn2 - 1;
			ndq = new RotDeque[nn2];
			for (i = 0; i < nn2; i++)
				ndq[i] = new RotDeque(nn2);

			// Copy the deque's contents
			for (i = 0; i < n2; i++)
			{
				deques[i].CopyTo(ndq[i / 2]);
				deques[i] = null;
			}

			// All done!
			n = nn;
			n2 = nn2;
			mask = nmask;
			next = next / 2;
			deques = ndq;
		}

		/// <summary>
		/// Shrink the tiered vector.
		/// </summary>
		/// <exception cref="System.InvalidOperationException">
		/// It would make no sense to shrink the vector, since it
		/// contains more than 2^(n-3) elements.
		/// </exception>
		protected virtual void Shrink()
		{
			int nn, nn2, nmask, nnext, i;
			RotDeque[] nd, where;

			if (next >= n2 / 8)
				throw (new InvalidOperationException(
				    "A tiered vector should only be shrunk when needed"));
			nn = n - 1;
			nn2 = 1 << nn;
			nmask = nn2 - 1;
			nd = new RotDeque[nn2];
			for (i = 0; i < nn2; i++)
				nd[i] = new RotDeque(nn2);
			where = new RotDeque[2];
			for (i = 0; i <= next; i++)
			{
				where[0] = nd[2*i];
				where[1] = nd[2*i + 1];
				deques[i].Split(where);
			}
			if (nd[2*i - 1].Empty)
				nnext = 2*i - 2;
			else
				nnext = 2*i - 1;

			n = nn;
			n2 = nn2;
			mask = nmask;
			next = nnext;
			deques = nd;
		}

		/// <summary>
		/// Get an enumerator on the tiered vector.
		/// </summary>
		/// <returns>A vector enumerator.</returns>
		public virtual IEnumerator GetEnumerator()
		{
			return (new TieredVectorEnumerator(this));
		}

		/// <summary>
		/// Copy the vector's elements to an array.
		/// </summary>
		/// <param name="array">The array to copy to.</param>
		/// <param name="idx">The index to start copying from.</param>
		/// <exception cref="System.ArgumentException">
		/// The <paramref name="array">array</paramref> parameter was
		/// either multidimensional or did not have enough free
		/// elements to accomodate all of the tiered vector's ones.
		/// </exception>
		/// <exception cref="System.ArgumentNullException">
		/// An undefined value was specified for the <paramref
		/// name="array">array</paramref> parameter.
		/// </exception>
		/// <exception cref="System.ArgumentOutOfRangeException">
		/// A negative value was specified for the <paramref
		/// name="idx">idx</paramref> parameter.
		/// </exception>
		public virtual void CopyTo(Array array, int idx)
		{
			int i;

			if (array == null)
				throw (new ArgumentNullException("array"));
			if (idx < 0)
				throw (new ArgumentOutOfRangeException("idx"));
			if (array.Rank != 1)
				throw (new ArgumentException(
				    "Cannot copy a TieredVector to a multidimensional array",
				    "array"));
			if (idx + Count > array.Length)
				throw (new ArgumentException(
				    "Not enough space left in the array to accomodate the TieredVector's elements",
				    "idx"));

			for (i = 0; i <= next; i++)
			{
				deques[i].CopyTo(array, idx);
				idx += deques[i].Count;
			}
		}

		/// <summary>
		/// Get a value indicating whether the tiered vector is safe
		/// for multithreaded operations.
		/// </summary>
		public virtual bool IsSynchronized
		{
			get
			{
				return (false);
			}
		}

		/// <summary>
		/// Get an object suitable for thread synchronization.
		/// </summary>
		public virtual object SyncRoot
		{
			get
			{
				return (this);
			}
		}

		/// <summary>
		/// Perform a shallow copy of the tiered vector.
		/// </summary>
		/// <returns>A shallow copy of the vector.</returns>
		public virtual object Clone()
		{
			TieredVector v;
			int i;

			v = new TieredVector(n);
			for (i = 0; i < n2; i++)
				deques[i].CopyTo(v.deques[i]);
			v.next = next;
			return (v);
		}

		/// <summary>
		/// An enumerator for the TieredVector class.
		/// </summary>
		[VCSAttribute("$Ringlet: cs/misc/RingletCollections/Ringlet.Collections/TieredVector.cs,v 1.6 2004/04/09 13:22:33 roam Exp $")]
		protected sealed class TieredVectorEnumerator : IEnumerator
		{
			/// <summary>
			/// The tiered vector to enumerate.
			/// </summary>
			private TieredVector v;
			/// <summary>
			/// The currently enumerated deque.
			/// </summary>
			private int d;
			/// <summary>
			/// The current deque enumerator.
			/// </summary>
			private IEnumerator en;

			/// <summary>
			/// Initialize a TieredVectorEnumerator object at
			/// the start of a tiered vector.
			/// </summary>
			/// <param name="vec">The vector to enumerate.</param>
			public TieredVectorEnumerator(TieredVector vec)
			{
				v = vec;
				Reset();
			}

			/// <summary>
			/// Reset the enumerator to the beginning of the
			/// vector.
			/// </summary>
			public void Reset()
			{
				d = 0;
				en = v.deques[d].GetEnumerator();
			}

			/// <summary>
			/// Retrieve the current element.
			/// </summary>
			public object Current
			{
				get
				{
					return (en.Current);
				}
			}

			/// <summary>
			/// Advance the enumerator.
			/// </summary>
			/// <returns>True if there are more elements.
			/// </returns>
			public bool MoveNext()
			{
				if (en.MoveNext())
					return (true);
				if (d == v.next)
					return (false);
				d++;
				en = v.deques[d].GetEnumerator();
				return (en.MoveNext());
			}
		}
	}
}
